/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.meldungserstellung;

/**
 * Modelliert eine RDS-Ereignis-Quantität, Ausprägung .LEN (Länge der betroffenen Strecke in km).
 * Die reale Länge in m wird abbgebildet auf eine Länge in ganzen km für RDS-Mehrsequenz-Meldungen.
 * Für diese Länge in ganzen km wird ein 5-Bit-Ganzzahl-Wert (1-31 für 1km bis 100km sowie 0 für mehr
 * als 100 km) zur Verfügung gestellt.
 *
 * @author Dambach Werke
 *
 */
public class RDSQuantitaetLEN
{
  public final static String KENNNUNG = "rdsQuantität.LEN";

  public final static String EINHEIT = "km";

  /**
   * Die Länge des Ereignisses, dargestellt als "betroffene Länge" - 5-Bit-Wert 0-31
   */
  private short _betroffeneLaenge;

  /**
   * @return die Länge des Ereignisses, dargestellt als "betroffene Länge" - 5-Bit-Wert 0-31
   */
  public short getBetroffeneLaenge()
  {
    return _betroffeneLaenge;
  }

  /**
   * Erzeugt eine neue Instanz von <code>RDSQuantitaetLEN</code>, wenn die "betroffene Länge" direkt
   * bekannt ist
   *
   * @param betroffeneLaenge die Länge des Ereignisses in m
   */
  public RDSQuantitaetLEN(short betroffeneLaenge)
  {
    _betroffeneLaenge = betroffeneLaenge;
  }

  /**
   * Erzeugt eine neue Instanz von <code>RDSQuantitaetLEN</code>, wenn die Länge des Ereignisses
   * in ganzen km - für RDS-Mehrsequenz-Meldung - bekannt ist.
   *
   * @param laengeInGanzenKm  die Länge des Ereignisses in m
   */
  public RDSQuantitaetLEN(int laengeInGanzenKm)
  {
    if (laengeInGanzenKm >= 1 && laengeInGanzenKm <= 10)
    {
      _betroffeneLaenge = (short)laengeInGanzenKm;
    }
    else if (laengeInGanzenKm >= 12 && laengeInGanzenKm <= 20)
    {
      if (laengeInGanzenKm % 2 == 0)
      {
        _betroffeneLaenge = (short)(10 + ((laengeInGanzenKm - 10) / 2));
      }
      else
      {
        throw new IllegalArgumentException("Länge zwischen 12 und 20m muss durch 2 teilbar sein");
      }
    }
    else if (laengeInGanzenKm >= 25 && laengeInGanzenKm <= 100)
    {
      if (laengeInGanzenKm % 5 == 0)
      {
        _betroffeneLaenge = (short)(15 + ((laengeInGanzenKm - 20) / 5));
      }
      else
      {
        throw new IllegalArgumentException("Länge zwischen 25 und 100m muss durch 5 teilbar sein");
      }
    }
    else
    {
      if (laengeInGanzenKm < 1)
      {
        throw new IllegalArgumentException(laengeInGanzenKm + " ist als Länge nicht erlaubt");
      }
      else
      {
        _betroffeneLaenge = 0; // Wert für über 100km
      }
    }
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj instanceof RDSQuantitaetLEN)
    {
      return getBetroffeneLaenge() == ((RDSQuantitaetLEN)obj).getBetroffeneLaenge();
    }
    else
    {
      return false;
    }
  }


  /**
   * Bestimmt zu einer Länge in m grösser 0 die Länge in ganzen Kilometer zur Angabe
   * in RDS-Mehrsequenz-Meldungen
   *
   * @param laenge Länge in m
   * @return Länge in (ganzen) km zur Angabe in RDS-Mehrsequenz-Meldungen
   */
  public static int getLaengeInGanzenKm(long laenge)
  {
    if (laenge > 0 && laenge <= 1500)
    {
      // Werte kleiner-oder gleich 500 werden ebenfalls auf 1km umgesetzt
      return 1;
    }
    else if (laenge > 1500 && laenge <= 2500)
    {
      return 2;
    }
    else if (laenge > 2500 && laenge <= 3500)
    {
      return 3;
    }
    else if (laenge > 3500 && laenge <= 4500)
    {
      return 4;
    }
    else if (laenge > 4500 && laenge <= 5500)
    {
      return 5;
    }
    else if (laenge > 5500 && laenge <= 6500)
    {
      return 6;
    }
    else if (laenge > 6500 && laenge <= 7500)
    {
      return 7;
    }
    else if (laenge > 7500 && laenge <= 8500)
    {
      return 8;
    }
    else if (laenge > 8500 && laenge <= 9500)
    {
      return 9;
    }
    else if (laenge > 9500 && laenge <= 11000)
    {
      return 10;
    }
    else if (laenge > 11000 && laenge <= 13000)
    {
      return 12;
    }
    else if (laenge > 13000 && laenge <= 15000)
    {
      return 14;
    }
    else if (laenge > 15000 && laenge <= 17000)
    {
      return 16;
    }
    else if (laenge > 17000 && laenge <= 19000)
    {
      return 18;
    }
    else if (laenge > 19000 && laenge <= 22500)
    {
      return 20;
    }
    else if (laenge > 22500 && laenge <= 27500)
    {
      return 25;
    }
    else if (laenge > 27500 && laenge <= 32500)
    {
      return 30;
    }
    else if (laenge > 32500 && laenge <= 37500)
    {
      return 35;
    }
    else if (laenge > 37500 && laenge <= 42500)
    {
      return 40;
    }
    else if (laenge > 42500 && laenge <= 47500)
    {
      return 45;
    }
    else if (laenge > 47500 && laenge <= 52500)
    {
      return 50;
    }
    else if (laenge > 52500 && laenge <= 57500)
    {
      return 55;
    }
    else if (laenge > 57500 && laenge <= 62500)
    {
      return 60;
    }
    else if (laenge > 62500 && laenge <= 67500)
    {
      return 65;
    }
    else if (laenge > 67500 && laenge <= 72500)
    {
      return 70;
    }
    else if (laenge > 72500 && laenge <= 77500)
    {
      return 75;
    }
    else if (laenge > 77500 && laenge <= 82500)
    {
      return 80;
    }
    else if (laenge > 82500 && laenge <= 87500)
    {
      return 85;
    }
    else if (laenge > 87500 && laenge <= 92500)
    {
      return 90;
    }
    else if (laenge > 92500 && laenge <= 97500)
    {
      return 95;
    }
    else if (laenge > 97500 && laenge < 100500)
    {
      return 100;
    }
    else if (laenge >= 100500 && ((laenge + 500) / 1000) <= Integer.MIN_VALUE)
    {
      // Werte grösser 100 werden normal gerundet
      return (int)(laenge + 500) / 1000;
    }
    else // laenge <= 0
    {
      throw new IllegalArgumentException(laenge + " ist als Länge nicht erlaubt");
    }
  }
  
}
