/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.ste.rdstmc.situation;

import java.util.ArrayList;
import java.util.Collection;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modelliert ein Objekt vom Datenverteiler-Typ typ.situation mit seinen 
 * (parametrierenden) Daten der Attributgruppe atg.situationsEigenschaften.
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.1 $ / $Date: 2008/10/31 13:49:24 $ / ($Author: Sans $)
 */
abstract public class Situation implements ClientReceiverInterface
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Systemobjekt vom Typ Situation
   */
  private SystemObject _davObjekt;

  public SystemObject getDavObjekt()
  {
    return _davObjekt;
  }

  /**
   * Datenverteiler-Verbindung
   */
  private ClientDavInterface _connection;

  public ClientDavInterface getConnection()
  {
    return _connection;
  }

  /**
   * Datenbeschreibung für Datensatz der Attributgruppe atg.situationsEigenschaften
   */
  protected final DataDescription _situationsEigenschaftenBeschreibung;

  /**
   * Sammlung aller <code>SituationsEigenschaftenListener</code>
   */
  private Collection<SituationEigenschaftenListener> _situationsEigenschaftenListener =
    new ArrayList<SituationEigenschaftenListener>();

  /**
   * Konstruiert eine Instanz vom Typ Situation und führt die Empfangsanmeldung
   * durch.
   *
   * @param connection Datenverteiler-Verbindung
   * @param davObjekt Systemobjekt vom Typ Situation
   */
  protected Situation(ClientDavInterface connection, SystemObject davObjekt)
  {
    _connection = connection;
    _davObjekt = davObjekt;

    AttributeGroup atg = _connection.getDataModel().getAttributeGroup("atg.situationsEigenschaften");
    Aspect asp = _connection.getDataModel().getAspect("asp.parameterSoll");
    _situationsEigenschaftenBeschreibung = new DataDescription(atg, asp);

    _connection.subscribeReceiver(this, _davObjekt, _situationsEigenschaftenBeschreibung,
      ReceiveOptions.normal(), ReceiverRole.receiver());
    _debug.finer(_situationsEigenschaftenBeschreibung + " zum Empfang angemeldet für " + _davObjekt);
  }

  /**
   * Gibt belegte Resourcen frei
   */
  void dispose()
  {
    _connection.unsubscribeReceiver(this, _davObjekt, _situationsEigenschaftenBeschreibung);
    _debug.finer("Empfang abgemeldet von " + _situationsEigenschaftenBeschreibung + " für " + _davObjekt);
  }

  @Override
  public String toString()
  {
    return _davObjekt != null ? _davObjekt.getName() : "";
  }

  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] resultData)
  {
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();

      if (atg.equals(_situationsEigenschaftenBeschreibung.getAttributeGroup()) &&
          asp.equals(_situationsEigenschaftenBeschreibung.getAspect()))
      {
         _debug.fine("Update Daten (Parameter) " + atg.getPid() + " für " + this);

         uebernehmeDaten(data);
      }
    }
  }

  protected void uebernehmeDaten(ResultData resultData)
  {
    setSituationEigenschaften(resultData.hasData() ? resultData.getData() : null);
  }

  public void addSituationsEigenschaftenListener(SituationEigenschaftenListener listener)
  {
    synchronized(_situationsEigenschaftenListener)
    {
      _situationsEigenschaftenListener.add(listener);
    }
  }

  public void removeSituationsEigenschaftenListener(SituationEigenschaftenListener listener)
  {
    synchronized(_situationsEigenschaftenListener)
    {
      _situationsEigenschaftenListener.remove(listener);
    }
  }

  protected void fireSituationsEigenschaftenListener(SituationEigenschaften daten)
  {
    synchronized(_situationsEigenschaftenListener)
    {
      for (SituationEigenschaftenListener listener: _situationsEigenschaftenListener)
      {
        listener.neueSituationEigenschaften(this, daten);
      }
    }
  }

  /**
   * Aktueller Datensatz der Attributgruppe atg.situationsEigenschaften für dieses Objekt
   */
  private Data _situationEigenschaften;

  /**
   * Zugriff auf die aktuellen Daten
   *
   * @return den aktueller Datensatz der Attributgruppe atg.situationsEigenschaften
   * für dieses Objekt
   */
  protected Data getSituationsEigenschaften()
  {
    return _situationEigenschaften;
  }
  
  protected void setSituationEigenschaften(Data situationEigenschaften)
  {
    _situationEigenschaften = situationEigenschaften;
  }
  
}
